'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.joinPrefix = joinPrefix;
exports.quotePrefix = quotePrefix;
exports.thisIsNotTheEndOfThisBatch = thisIsNotTheEndOfThisBatch;
exports.whereConditionIsntSupposedToGoInsideSubqueryOrOnNextBatch = whereConditionIsntSupposedToGoInsideSubqueryOrOnNextBatch;
exports.keysetPagingSelect = keysetPagingSelect;
exports.offsetPagingSelect = offsetPagingSelect;
exports.orderColumnsToString = orderColumnsToString;
exports.interpretForOffsetPaging = interpretForOffsetPaging;
exports.interpretForKeysetPaging = interpretForKeysetPaging;
exports.validateCursor = validateCursor;

var _lodash = require('lodash');

var _graphqlRelay = require('graphql-relay');

var _util = require('../util');

function joinPrefix(prefix) {
  return prefix.slice(1).map(name => name + '__').join('');
}

function doubleQuote(str) {
  return `"${str}"`;
}

function quotePrefix(prefix, q = doubleQuote) {
  return prefix.map(name => q(name));
}

function thisIsNotTheEndOfThisBatch(node, parent) {
  return !node.sqlBatch && !(node.junction && node.junction.sqlBatch) || !parent;
}

function whereConditionIsntSupposedToGoInsideSubqueryOrOnNextBatch(node, parent) {
  return !node.paginate && (!(node.sqlBatch || node.junction && node.junction.sqlBatch) || !parent);
}

function keysetPagingSelect(table, whereCondition, order, limit, as, options = {}) {
  let { joinCondition, joinType, extraJoin, q } = options;
  q = q || doubleQuote;
  whereCondition = (0, _lodash.filter)(whereCondition).join(' AND ') || 'TRUE';
  if (joinCondition) {
    return `\
${joinType || ''} JOIN LATERAL (
  SELECT ${q(as)}.*
  FROM ${table} ${q(as)}
  ${extraJoin ? `LEFT JOIN ${extraJoin.name} ${q(extraJoin.as)}
    ON ${extraJoin.condition}` : ''}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(order.columns, q, order.table)}
  LIMIT ${limit}
) ${q(as)} ON ${joinCondition}`;
  } else {
    return `\
FROM (
  SELECT ${q(as)}.*
  FROM ${table} ${q(as)}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(order.columns, q, order.table)}
  LIMIT ${limit}
) ${q(as)}`;
  }
}

function offsetPagingSelect(table, pagingWhereConditions, order, limit, offset, as, options = {}) {
  let { joinCondition, joinType, extraJoin, q } = options;
  q = q || doubleQuote;
  const whereCondition = (0, _lodash.filter)(pagingWhereConditions).join(' AND ') || 'TRUE';
  if (joinCondition) {
    return `\
${joinType || ''} JOIN LATERAL (
  SELECT ${q(as)}.*, count(*) OVER () AS ${q('$total')}
  FROM ${table} ${q(as)}
  ${extraJoin ? `LEFT JOIN ${extraJoin.name} ${q(extraJoin.as)}
    ON ${extraJoin.condition}` : ''}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(order.columns, q, order.table)}
  LIMIT ${limit} OFFSET ${offset}
) ${q(as)} ON ${joinCondition}`;
  } else {
    return `\
FROM (
  SELECT ${q(as)}.*, count(*) OVER () AS ${q('$total')}
  FROM ${table} ${q(as)}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(order.columns, q, order.table)}
  LIMIT ${limit} OFFSET ${offset}
) ${q(as)}`;
  }
}

function orderColumnsToString(orderColumns, q, as) {
  const conditions = [];
  for (let column in orderColumns) {
    conditions.push(`${as ? q(as) + '.' : ''}${q(column)} ${orderColumns[column]}`);
  }
  return conditions.join(', ');
}

function interpretForOffsetPaging(node, dialect) {
  const { name } = dialect;
  if (node.args && node.args.last) {
    throw new Error('Backward pagination not supported with offsets. Consider using keyset pagination instead');
  }

  const order = {};
  if (node.orderBy) {
    order.table = node.as;
    order.columns = node.orderBy;
  } else {
    order.table = node.junction.as;
    order.columns = node.junction.orderBy;
  }

  let limit = ['mariadb', 'mysql', 'oracle'].includes(name) ? '18446744073709551615' : 'ALL';
  let offset = 0;
  if (node.args && node.args.first) {
    limit = parseInt(node.args.first);

    if (node.paginate) {
      limit++;
    }
    if (node.args.after) {
      offset = (0, _graphqlRelay.cursorToOffset)(node.args.after) + 1;
    }
  }
  return { limit, offset, order };
}

function interpretForKeysetPaging(node, dialect) {
  const { name } = dialect;

  let sortTable;
  let sortKey;
  let descending;
  const order = { columns: {} };
  if (node.sortKey) {
    sortKey = node.sortKey;
    descending = sortKey.order.toUpperCase() === 'DESC';
    sortTable = node.as;

    if (node.args && node.args.last) {
      descending = !descending;
    }
    for (let column of (0, _util.wrap)(sortKey.key)) {
      order.columns[column] = descending ? 'DESC' : 'ASC';
    }
    order.table = node.as;
  } else {
    sortKey = node.junction.sortKey;
    descending = sortKey.order.toUpperCase() === 'DESC';
    sortTable = node.junction.as;

    if (node.args && node.args.last) {
      descending = !descending;
    }
    for (let column of (0, _util.wrap)(sortKey.key)) {
      order.columns[column] = descending ? 'DESC' : 'ASC';
    }
    order.table = node.junction.as;
  }

  let limit = ['mariadb', 'mysql', 'oracle'].includes(name) ? '18446744073709551615' : 'ALL';
  let whereCondition = '';
  if (node.args && node.args.first) {
    limit = parseInt(node.args.first) + 1;
    if (node.args.after) {
      const cursorObj = (0, _util.cursorToObj)(node.args.after);
      validateCursor(cursorObj, (0, _util.wrap)(sortKey.key));
      whereCondition = sortKeyToWhereCondition(cursorObj, descending, sortTable, dialect);
    }
    if (node.args.before) {
      throw new Error('Using "before" with "first" is nonsensical.');
    }
  } else if (node.args && node.args.last) {
    limit = parseInt(node.args.last) + 1;
    if (node.args.before) {
      const cursorObj = (0, _util.cursorToObj)(node.args.before);
      validateCursor(cursorObj, (0, _util.wrap)(sortKey.key));
      whereCondition = sortKeyToWhereCondition(cursorObj, descending, sortTable, dialect);
    }
    if (node.args.after) {
      throw new Error('Using "after" with "last" is nonsensical.');
    }
  }

  return { limit, order, whereCondition };
}

function validateCursor(cursorObj, expectedKeys) {
  const actualKeys = Object.keys(cursorObj);
  const expectedKeySet = new Set(expectedKeys);
  const actualKeySet = new Set(actualKeys);
  for (let key of actualKeys) {
    if (!expectedKeySet.has(key)) {
      throw new Error(`Invalid cursor. The column "${key}" is not in the sort key.`);
    }
  }
  for (let key of expectedKeys) {
    if (!actualKeySet.has(key)) {
      throw new Error(`Invalid cursor. The column "${key}" is not in the cursor.`);
    }
  }
}

function sortKeyToWhereCondition(keyObj, descending, sortTable, dialect) {
  const { name, quote: q } = dialect;
  const sortColumns = [];
  const sortValues = [];
  for (let key in keyObj) {
    sortColumns.push(`${q(sortTable)}.${q(key)}`);
    sortValues.push((0, _util.maybeQuote)(keyObj[key], name));
  }
  const operator = descending ? '<' : '>';
  return name === 'oracle' ? recursiveWhereJoin(sortColumns, sortValues, operator) : `(${sortColumns.join(', ')}) ${operator} (${sortValues.join(', ')})`;
}

function recursiveWhereJoin(columns, values, op) {
  const condition = `${columns.pop()} ${op} ${values.pop()}`;
  return _recursiveWhereJoin(columns, values, op, condition);
}

function _recursiveWhereJoin(columns, values, op, condition) {
  if (!columns.length) {
    return condition;
  }
  const column = columns.pop();
  const value = values.pop();
  condition = `(${column} ${op} ${value} OR (${column} = ${value} AND ${condition}))`;
  return _recursiveWhereJoin(columns, values, op, condition);
}